from datetime import datetime, timezone
from glob import iglob
import logging
import os
from time import sleep
from typing import Any, Dict, List, Optional

import tator
from tator.openapi.tator_openapi import CreateListResponse, File, FileType, Media, MediaType, TatorApi
from tqdm import tqdm

from hms_import.b3.generate_states import generate_multis_and_states
from hms_import.util import validate_types
from hms_import.summary_image import create_summary_image

logger = logging.getLogger(__name__)
DATETIME_FORMAT = "%Y%m%dT%H%M%SZ"


def upload_and_import_videos(
    tator_api: TatorApi,
    video_list: List[str],
    media_type: MediaType,
    attrs: Dict[str, Any],
    file_list: List[File],
    section_name: str,
) -> List[int]:
    """Returns list of created media ids"""
    created_ids = []
    media_type_id = media_type.id
    project_id = media_type.project
    shared_attributes = {
        **attrs,
        "related_files": ",".join(str(file_obj.id) for file_obj in file_list),
    }

    for video_path in tqdm(video_list, desc="Video Imports", dynamic_ncols=True, position=0):
        # Initialize attributes with sentinel values
        try:
            filename = os.path.basename(video_path)
            filename_parts = os.path.splitext(filename)[0].split("-")
            start_datetime = datetime.min
            if len(filename_parts) == 4:
                start_date, start_time = filename_parts[2:4]
                try:
                    start_datetime = datetime.strptime(
                        f"{start_date}T{start_time}", DATETIME_FORMAT
                    ).replace(tzinfo=timezone.utc)
                except Exception:
                    logger.warning(
                        "Could not parse datetime from filename '%s'", filename, exc_info=True
                    )
            attributes = {**shared_attributes, "toc_start": start_datetime}
            vid_md5 = tator.util.md5sum(video_path)
            media_spec = {
                "attributes": attributes,
                "name": filename,
                "type": media_type_id,
                "section": section_name,
                "md5": vid_md5,
            }
            response = tator_api.create_media_list(project_id, body=[media_spec])
            if isinstance(response, CreateListResponse) and response.id:
                created_ids.append(response.id[0])
                pbar = tqdm(desc=filename, dynamic_ncols=True, position=1, total=100)
                last_progress = 0
                for p, _ in tator.util.upload_media(
                    api=tator_api,
                    type_id=media_type_id,
                    path=video_path,
                    md5=vid_md5,
                    section=section_name,
                    fname=filename,
                    attributes=attrs,
                    media_id=created_ids[-1],
                    timeout=120,
                ):
                    increment = max(p - last_progress, 1)
                    pbar.update(increment)
                    last_progress = p

            else:
                logger.warning("Could not create media object for '%s', skipping", video_path)
        except Exception:
            logger.error(
                "Encountered exception while processing '%s', skipping", video_path, exc_info=True
            )

    return created_ids


def upload_sensor_data(
    tator_api: TatorApi,
    sensor_list: List[str],
    file_type: FileType,
) -> List[File]:
    file_type_id = file_type.id
    file_list = []
    for sensor_path in tqdm(sensor_list, dynamic_ncols=True, desc="Sensor Import", position=0):
        filename = os.path.basename(sensor_path)
        try:
            pbar = tqdm(desc=filename, dynamic_ncols=True, position=1, total=100)
            last_progress = 0
            response = None
            for p, response in tator.util.upload_generic_file(
                api=tator_api,
                file_type=file_type_id,
                path=sensor_path,
                description="Raw sensor data",
                name=filename,
                timeout=120,
            ):
                increment = max(p - last_progress, 1)
                pbar.update(increment)
                last_progress = p
        except Exception:
            logger.warning(
                "Encountered exception while processing '%s', skipping", sensor_path, exc_info=True
            )
            continue
        if isinstance(response, File):
            file_list.append(response)
    return file_list


def main(
    host: str,
    token: str,
    media_type_id: int,
    file_type_id: int,
    multi_type_id: int,
    state_type_id: int,
    image_type_id: int,
    directory: str,
    sail_date: Optional[str] = None,
    land_date: Optional[str] = None,
    hdd_recv_date: Optional[str] = None,
    hdd_sn: Optional[str] = None,
) -> bool:
    """
    :host: The Tator domain
    :token: The REST API token
    :media_type_id: The unique ID of the type of video to create
    :file_type_id: The unique ID of the type of file to create for storing GPS data
    :multi_type_id: The unique ID of the type of multiview to create
    :state_type_id: The unique ID of the type of State to create
    :image_type_id: The unique ID of the type of summary image to create
    :directory: The folder containing the files to import
    :sail_date: The sail date of the trip
    :land_date: The land date of the trip
    :hdd_recv_date: The date the hard drive was received
    :hdd_sn: The hard drive serial number
    :returns: True if the import was successful, False if any part of it fails
    """

    # Validate the given media and file types, abort if they do not exist or are incompatible
    tator_api = tator.get_api(host=host, token=token)
    media_type, file_type = validate_types(tator_api, media_type_id, file_type_id)
    if media_type is None:
        logger.error("Could not get media type %d from Tator, aborting", media_type_id)
        return False
    if file_type is None:
        logger.error("Could not get file type %d from Tator, aborting", file_type_id)
        return False
    if media_type.project != file_type.project:
        logger.error(
            "Received MediaType %d and FileType %d, which are from different projects, aborting",
            media_type_id,
            file_type_id,
        )
        return False

    # Locate media for import and create section name
    video_list = list(iglob(os.path.join(directory, f"*.mp4")))
    video_path = video_list[0]
    filename = os.path.basename(video_path)
    filename_parts = os.path.splitext(filename)[0].split("-")
    section_name = f"Imported on {datetime.now()}"
    vessel_name = "UNKNOWN"
    if len(filename_parts) == 4:
        try:
            vessel_name = filename_parts[0]
            section_name = f"{vessel_name} - ({sail_date})"
        except Exception:
            logger.warning("Could not get the vessel or section name", exc_info=True)
    section_name, shared_attrs = create_summary_image(
        host=host,
        token=token,
        media_type=image_type_id,
        import_type="B3",
        directory=directory,
        sail_date=sail_date,
        land_date=land_date,
        hdd_recv_date=hdd_recv_date,
        hdd_sn=hdd_sn,
    )

    # Locate sensor data for import
    sensor_list = list(iglob(os.path.join(directory, f"*.log")))
    if sensor_list:
        file_list = upload_sensor_data(tator_api, sensor_list, file_type)
    else:
        file_list = []
        logger.warning("No sensor data found, only videos will be imported")

    if video_list:
        created_ids = upload_and_import_videos(
            tator_api, video_list, media_type, shared_attrs, file_list, section_name
        )
    else:
        logger.error("No media found, aborting")
        return False

    # Wait for imports to generate thumbnails
    logger.info("Waiting for transcodes to generate thumbnails for all uploads...")
    pbar = tqdm(total=len(created_ids), dynamic_ncols=True)
    while created_ids:
        media = None
        media_id = created_ids.pop(0)
        try:
            media = tator_api.get_media(media_id)
        except Exception:
            logger.warning("Could not get media '%d'", media_id, exc_info=True)
        if not isinstance(media, Media) or (media.media_files and media.media_files.thumbnail_gif):
            # Do not put the id back on the list if it is invalid or it is ready; check the next one
            pbar.update()
            continue

        # Put the id at the end to check later and sleep to avoid a tight loop
        created_ids.append(media_id)
        sleep(1.0)

    # Generate associated multiviews and GPS States
    generate_multis_and_states(
        host=host,
        token=token,
        media_type_id=media_type_id,
        multi_type_id=multi_type_id,
        state_type_id=state_type_id,
        section_name=section_name,
    )
    return True



from argparse import ArgumentParser, RawDescriptionHelpFormatter
import logging
from logging.handlers import TimedRotatingFileHandler
import os

import tator

# Log info and up to console, everything to file
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
console_handler = logging.StreamHandler()
console_handler.setLevel(logging.INFO)
file_handler = TimedRotatingFileHandler(
    os.path.join(os.getcwd(), f"{__name__.split('.')[0]}.log"), when="midnight", backupCount=7
)
file_handler.setLevel(logging.DEBUG)
formatter = logging.Formatter("[%(asctime)s - %(levelname)s - %(name)s]: %(message)s")
console_handler.setFormatter(formatter)
file_handler.setFormatter(formatter)
logger.addHandler(console_handler)
logger.addHandler(file_handler)


def get_parser():
    parser = ArgumentParser(
        description="Script for importing video and metadata in O2 and B3 formats.",
        formatter_class=RawDescriptionHelpFormatter,
    )
    cmd_parser = parser.add_subparsers(title="Commands", dest="command")

    upload_parser = cmd_parser.add_parser(
        "b3-upload", help="Imports video and GPS files from unlocked LUKS-encrypted device"
    )
    upload_parser = tator.get_parser(upload_parser)
    upload_parser.add_argument("--media-type-id", type=int, help="")
    upload_parser.add_argument("--file-type-id", type=int, help="")
    upload_parser.add_argument("--multi-type-id", type=int, help="")
    upload_parser.add_argument("--state-type-id", type=int, help="")
    upload_parser.add_argument("--image-type-id", type=int, help="")
    upload_parser.add_argument("--directory", type=str, help="")
    upload_parser.add_argument("--sail-date", type=str, help="", required=False)
    upload_parser.add_argument("--land-date", type=str, help="", required=False)
    upload_parser.add_argument("--hdd-recv-date", type=str, help="", required=False)
    upload_parser.add_argument("--hdd-sn", type=str, help="", required=False)

    import_parser = cmd_parser.add_parser(
        "o2-upload", help="Script for uploading raw, encrypted video files."
    )
    import_parser.add_argument(
        "config_file", type=str, help=f"The configuration .ini file used to initialize {__name__}."
    )
    return parser


def main() -> None:
    # Parse arguments
    parser = get_parser()
    args = parser.parse_args()
    argdict = vars(args)
    command = argdict.pop("command")

    # Import the desired main function
    if command == "o2-upload":
        from hms_import.o2 import main
    elif command == "b3-upload":
        from hms_import.b3 import main
    else:
        raise RuntimeError(f"Received unhandled command '{command}'")

    main(**argdict)



import logging
from typing import Any, Callable, Optional, Tuple

from tator.openapi.tator_openapi import FileType, MediaType, TatorApi

logger = logging.getLogger(__name__)


def validate_type_from_config(api, config, config_field, project_id, tator_type, list_filters=None):
    if tator_type not in ["media_type", "file_type"]:
        raise ValueError(f"Cannot validate {config_field=} for {tator_type=}")

    type_id = config["Tator"].getint(config_field, -1)
    if type_id > 0:
        try:
            getattr(api, f"get_{tator_type}")(type_id)
        except Exception as exc:
            raise ValueError(f"Could not find {config_field} with id {type_id}") from exc
    else:
        try:
            types = getattr(api, f"get_{tator_type}_list")(project_id)
        except Exception as exc:
            raise RuntimeError(f"Could not list {config_field}s from project {project_id}") from exc
        if list_filters:
            for attr, value in list_filters:
                types = [
                    _type
                    for _type in types
                    if hasattr(_type, attr) and getattr(_type, attr) == value
                ]
        if len(types) > 1:
            raise ValueError(
                f"Project {project_id} has more than one {config_field}, specify one of the "
                f"following in the config: {types}"
            )
        type_id = types[0].id
    return type_id


def safe_get_type(type_id: int, type_getter: Callable[[int], Any]):
    type_inst = None
    try:
        type_inst = type_getter(type_id)
    except Exception:
        logger.error("Could not find type %d", type_id, exc_info=True)
    return type_inst


def validate_types(
    tator_api: TatorApi, media_type_id: int, file_type_id: int
) -> Tuple[Optional[MediaType], Optional[FileType]]:
    """Validates media and file types by ensuring they both exist and belong to the same project."""
    return (
        safe_get_type(media_type_id, tator_api.get_media_type),
        safe_get_type(file_type_id, tator_api.get_file_type),
    )



from collections import defaultdict
import datetime
import logging
from math import floor
import os
import requests
import tempfile

from natsort import natsorted
import pynmea2
import tator
from tator.openapi.tator_openapi import File, MediaType
from tqdm import tqdm

DATETIME_FORMAT = "%Y%m%dT%H%M%SZ"
logger = logging.getLogger(__file__)


def _line_to_attrs(line):
    try:
        message = pynmea2.parse(f"$GPRMC,{line}")
        return {
            "Knots": message.spd_over_grnd,
            "Heading": message.true_course,
            "Datecode": message.datetime,
            "Position": [message.lon, message.lat],
        }
    except Exception:
        logger.warning(f"Could not parse gps entry: '{line}'", exc_info=True)
    return None


def points_from_file(gps_data_path):
    with open(gps_data_path, "r") as fp:
        lines = fp.readlines()
    metadata = [_line_to_attrs(line) for line in lines]
    return [point for point in metadata if point is not None]


def generate_multis_and_states(
    host: str,
    token: str,
    media_type_id: int,
    multi_type_id: int,
    state_type_id: int,
    section_name: str,
):
    tator_api = tator.get_api(host=host, token=token)

    section = None
    try:
        multi_type = tator_api.get_media_type(multi_type_id)
    except Exception:
        logger.error(
            "Could not find MediaType '%d' in Tator, aborting!", multi_type_id, exc_info=True
        )
        return
    if not isinstance(multi_type, MediaType):
        return
    project_id = multi_type.project

    try:
        section_list = tator_api.get_section_list(project_id)
    except Exception:
        logger.error(
            "Could not find MediaType '%d' in Tator, aborting!", multi_type_id, exc_info=True
        )
        return

    if isinstance(section_list, list):
        for sec in section_list:
            if sec.name == section_name:
                section = sec
                break
    if not section:
        logger.error(
            "Found no section named '%s' in project '%d', aborting!", section_name, project_id
        )
        return

    media_list = tator_api.get_media_list(
        project=project_id, section=section.id, type=media_type_id
    )
    if not (isinstance(media_list, list) or media_list):
        logger.error(
            "Found no media in section '%s' in project '%d', aborting!", section_name, project_id
        )
        return

    project_id = media_list[0].project
    datetime_lookup = {}
    multi_lookup = defaultdict(list)
    state_file_ids = set()

    for media in media_list:
        try:
            # Get info to create states
            state_file_ids.update(
                int(fid)
                for fid in media.attributes.get("related_files", "").split(",")
                if fid.isnumeric()
            )

            toc_start = datetime.datetime.fromisoformat(media.attributes["toc_start"])
            duration = datetime.timedelta(seconds=media.num_frames / media.fps)
            datetime_lookup[media.id] = {
                "fps": media.fps,
                "toc_start": toc_start,
                "toc_end": toc_start + duration,
            }

            # Get info to create multiviews
            if not media.name:
                logger.warning("Could not get filename for media '%d', skipping", media.id)
                continue
            filename = media.name
            multi_lookup[toc_start].append((filename, media.id))
        except Exception:
            logger.error(
                "Encountered exception while processing '%d', skipping", media.id, exc_info=True
            )

    # Make multiviews from pairs of concurrent videos
    for start_dt, media_lst in tqdm(
        list(multi_lookup.items()), desc="Creating Multis", dynamic_ncols=True, position=0
    ):
        sorted_ids = [ele[-1] for ele in natsorted(media_lst)]
        tator.util.make_multi_stream(
            api=tator_api,
            type_id=multi_type_id,
            layout=[1, 2],
            name=start_dt.strftime(DATETIME_FORMAT),
            media_ids=sorted_ids,
            section=section.name,
        )

    # Generate states
    state_specs = []
    for file_id in state_file_ids:
        try:
            file_obj = tator_api.get_file(file_id)
        except Exception:
            logger.error("Could not get File '%d', skipping", file_id, exc_info=True)
            continue
        if not isinstance(file_obj, File):
            logger.warning("Could not find file with id %d, skipping", file_id)
            continue
        if not isinstance(file_obj.name, str):
            logger.warning("Could not get name of file %d, skipping", file_id)
            continue
        download_info = tator_api.get_download_info(
            file_obj.project, download_info_spec={"keys": [file_obj.path]}
        )

        if not (download_info and isinstance(download_info, list) and download_info[0].url):
            logger.warning(
                "Could not get download info for '%s' (%d), skipping", file_obj.name, file_id
            )
            continue
        else:
            download_url = download_info[0].url

        # Make working directories
        with tempfile.TemporaryDirectory() as data_dir:
            # Download the encrypted GPS file
            gps_data_path = os.path.join(data_dir, file_obj.name)
            response = requests.get(download_url, stream=True)
            with open(gps_data_path, "wb") as fp:
                for chunk in response.iter_content(chunk_size=128):
                    fp.write(chunk)

            metadata = points_from_file(gps_data_path)

        # Create a state spec for each point/media pair, where the point timestamp occurs between
        # the start and stop timestamps of the media object
        for point in metadata:
            for media_id, media_info in datetime_lookup.items():
                if (
                    media_info["toc_start"] <= point["Datecode"] <= media_info["toc_end"]
                    and media_info["fps"]
                ):
                    delta = (point["Datecode"] - media_info["toc_start"]).total_seconds()
                    frame = floor(delta * media_info["fps"])
                    state_specs.append(
                        {
                            "type": state_type_id,
                            "media_ids": [media_id],
                            "frame": frame,
                            "attributes": point,
                        }
                    )

    for response in tator.util.chunked_create(
        tator_api.create_state_list, project_id, body=state_specs
    ):
        logger.info(response.message)
